using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Runtime.ConstrainedExecution;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows;
using Microsoft.Win32.SafeHandles;
using XenAdmin.Core;

namespace XenAdmin.Native
{
	public delegate IntPtr OFNHookProcDelegate(IntPtr hdlg, int msg, IntPtr wParam, IntPtr lParam);
    public delegate bool EnumWindowsProc(IntPtr window, int i);
    public delegate bool EnumUILanguagesProc(string lpUILanguageString, IntPtr lParam);
    public delegate int LowLevelKeyboardProc(int nCode, int wParam, ref KBDLLHOOKSTRUCT lParam);

	[StructLayout(LayoutKind.Sequential)]
    public struct ICONINFO
    {
        public bool fIcon;
        public int xHotspot;
        public int yHotspot;
        public IntPtr hbmMask;
        public IntPtr hbmColor;
    }

	[StructLayout(LayoutKind.Sequential)]
	public struct POINT
    {
        public int X;
        public int Y;
    }

	[StructLayout(LayoutKind.Sequential)]
    public struct RECT
    {
        public int left;
        public int top;
        public int right;
        public int bottom;
    }

	[StructLayout(LayoutKind.Sequential)]
    public struct COMBOBOXINFO
    {
        public int cbSize;
        public RECT rcItem;
        public RECT rcButton;
        public IntPtr stateButton;
        public IntPtr hwndCombo;
        public IntPtr hwndItem;
        public IntPtr hwndList;
    }

	[StructLayout(LayoutKind.Sequential)]
    public struct ScrollInfo
    {
        public uint cbSize;
        public uint fMask;
        public int nMin;
        public int nMax;
        public uint nPage;
        public int nPos;
        public int nTrackPos;
    }

	[StructLayout(LayoutKind.Sequential)]
    public struct NativeMessage
    {
        public IntPtr handle;
        public uint msg;
        public IntPtr wParam;
        public IntPtr lParam;
        public uint time;
		public Point p;
    }

	[StructLayout(LayoutKind.Sequential)]
    public struct FLASHWINFO
    {
        public UInt32 cbSize;
        public IntPtr hwnd;
        public UInt32 dwFlags;
        public UInt32 uCount;
        public UInt32 dwTimeout;
    }

	[StructLayout(LayoutKind.Sequential)]
	public struct KBDLLHOOKSTRUCT
    {
        public int vkCode;
        public int scanCode;
        public int flags;
        public int time;
        public int dwExtraInfo;
    }

	[StructLayoutAttribute(LayoutKind.Sequential)]
    public struct PROCESSENTRY32
    {
        public uint dwSize;
        public uint cntUsage;
        public uint th32ProcessID;
        public IntPtr th32DefaultHeapID;
        public uint th32ModuleID;
        public uint cntThreads;
        public uint th32ParentProcessID;
        public int pcPriClassBase;
        public uint dwFlags;

        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
        public string szExeFile;
    }

	[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
    public struct OPENFILENAME
    {
        public int lStructSize;
        public IntPtr hwndOwner;
        public int hInstance;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrFilter;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrCustomFilter;
        public int nMaxCustFilter;
        public int nFilterIndex;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrFile;
        public int nMaxFile;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrFileTitle;
        public int nMaxFileTitle;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrInitialDir;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrTitle;
        public int Flags;
        public short nFileOffset;
        public short nFileExtension;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpstrDefExt;
        public int lCustData;
        public OFNHookProcDelegate lpfnHook;
        [MarshalAs(UnmanagedType.LPTStr)]
        public string lpTemplateName;
        //only if on nt 5.0 or higher
        public int pvReserved;
        public int dwReserved;
        public int FlagsEx;
    }

	[Flags]
    public enum TOKEN_ACCESS : uint
    {
        TOKEN_QUERY = 0x0008
    }

	[Flags]
	public enum SnapshotFlags : uint
	{
		HeapList = 0x00000001,
		Process = 0x00000002,
		Thread = 0x00000004,
		Module = 0x00000008,
		Module32 = 0x00000010,
		Inherit = 0x80000000,
		All = 0x0000001F
	}

	public class ToolHelpHandle : SafeHandleZeroOrMinusOneIsInvalid
    {
        public ToolHelpHandle() : base(true)
        {
        }
    
        [ReliabilityContract(Consistency.WillNotCorruptState, Cer.MayFail)]
        override protected bool ReleaseHandle()
        {
			// TODO: return CloseHandle(handle);
			return false;
        }
    }

    public interface INativeCall
    {

        bool BringWindowToTop(IntPtr window);

        bool CloseWindow(IntPtr window);

        IntPtr CreateIconIndirect(ref ICONINFO iconinfo);

        bool DestroyIcon(IntPtr hIcon);

        IntPtr SendMessage(IntPtr hWnd, uint Msg, uint wParam, string lParam);

        bool GetComboBoxInfo(IntPtr hwnd, ref COMBOBOXINFO pcbi);

        bool SetForegroundWindow(IntPtr window);

        bool OpenIcon(IntPtr window);

        IntPtr FindWindowEx(IntPtr parent, IntPtr childAfter, string className, string windowName);

        IntPtr FindWindowWin32(string className, string windowName);

        int SendMessage(IntPtr window, int message, int wparam, int lparam);

        int PostMessage(IntPtr window, int message, int wparam, int lparam);

        IntPtr GetParent(IntPtr window);

        IntPtr GetDesktopWindow();

        IntPtr GetForegroundWindow();

        IntPtr GetLastActivePopup(IntPtr window);

        int GetWindowText(IntPtr window, ref StringBuilder text, int copyCount);

        bool SetWindowText(IntPtr window, string text);

        int GetWindowTextLength(IntPtr window);

        int SetWindowLong(IntPtr window, int index, int value);

		IntPtr GetWindowLong(IntPtr window, int index);

		int GetClassName(IntPtr window, ref StringBuilder text, int copyCount);

        bool EnumChildWindows(IntPtr window, EnumWindowsProc callback, int i);

        bool EnumThreadWindows(int threadId, EnumWindowsProc callback, int i);

        bool EnumWindows(EnumWindowsProc callback, int i);

        int GetWindowThreadProcessId(IntPtr window, ref int processId);

        int GetWindowThreadProcessId(IntPtr window, IntPtr ptr);

        bool IsChild(IntPtr parent, IntPtr window);

        bool IsIconic(IntPtr window);

        bool IsZoomed(IntPtr window);

        IntPtr GetWindowDC(IntPtr hwnd);

        int ReleaseDC(IntPtr hwnd, IntPtr dc);

        bool MoveWindow(IntPtr hwnd, int X, int Y, int width, int height, bool repaint);

        IntPtr GetNextWindow(IntPtr hwnd, int wFlag);

        IntPtr GetWindow(IntPtr hWnd, uint uCmd);

        bool PeekMessage(out NativeMessage message, IntPtr handle, uint filterMin, uint filterMax, uint removeMsg);

		IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wParam, ref POINT pt);

        IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam);

        IntPtr GetDCEx(IntPtr hWnd, IntPtr hrgnClip, int flags);

        int GetGuiResources(IntPtr hProcess, int uiFlags);

        int ToUnicode(uint wVirtKey, uint wScanCode, IntPtr lpKeyState, out StringBuilder pwszBuff, int cchBuff, uint wFlags);

        bool GetKeyboardState(IntPtr lpKeyState);

        Int32 FlashWindowEx(ref FLASHWINFO pwfi);

        IntPtr GetClipboardViewer();

        IntPtr SetClipboardViewer(IntPtr hWnd);

        bool ChangeClipboardChain(IntPtr hWnd, IntPtr hWndNext);

        IntPtr GetDlgItem(IntPtr hDlg, int nIDDlgItem);

        IntPtr CreateWindowEx(uint dwExStyle, string lpClassName, string lpWindowName, uint dwStyle, int x, int y, int nWidth, int nHeight, IntPtr hWndParent, IntPtr hMenu, IntPtr hInstance, IntPtr lpParam);

        bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int X, int Y, int cx, int cy, uint uFlags);

        bool ScreenToClient(IntPtr hWnd, ref POINT lpPoint);

        bool GetWindowRect(IntPtr hWnd, ref RECT lpRect);

        bool SetWindowText(int hWnd, string lpString);

        bool DestroyWindow(IntPtr hwnd);

        IntPtr SendMessage(IntPtr hWnd, int Msg, IntPtr wParam, string lParam);

        IntPtr SetFocus(IntPtr hwnd);

        bool GetScrollInfo(IntPtr hWnd, int n, ref ScrollInfo lpScrollInfo);

        int SetScrollInfo(IntPtr hwnd, int fnBar, ref ScrollInfo lpsi, bool fRedraw);

        IntPtr FindWindow(string lpClassName, string lpWindowName);

        bool PostMessage(IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam);

        int SetWindowLong(IntPtr hWnd, int nIndex, IntPtr newWndProc);

        IntPtr CallWindowProc(IntPtr lpPrevWndFunc, IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam);

        bool GetClientRect(IntPtr hWnd, out RECT lpRect);

        IntPtr SetWindowsHookEx(int idHook, LowLevelKeyboardProc lpfn, IntPtr hMod, uint dwThreadId);

        bool UnhookWindowsHookEx(IntPtr hhk);

        int CallNextHookEx(IntPtr hhk, int nCode, int wParam, ref KBDLLHOOKSTRUCT lParam);

        bool QueryPerformanceCounter(out long x);

        bool QueryPerformanceFrequency(out long x);

        bool Process32First(ToolHelpHandle hSnapshot, ref PROCESSENTRY32 lppe);

        bool Process32Next(ToolHelpHandle hSnapshot, ref PROCESSENTRY32 lppe);

        uint FormatMessage(uint dwFlags, IntPtr lpSource, uint dwMessageId, uint dwLanguageId, ref IntPtr lpBuffer, uint nSize, IntPtr pArguments);

        IntPtr LocalFree(IntPtr hMem);

        bool EnumUILanguages(EnumUILanguagesProc pUILanguageEnumProc, uint dwFlags, IntPtr lParam);

        bool GetDiskFreeSpaceEx(string lpszPath, ref long lpFreeBytesAvailable, ref long lpTotalNumberOfBytes, ref long lpTotalNumberOfFreeBytes);

        int GetCurrentThreadId();

		IntPtr CreateNamedPipe(String pipeName, uint openMode, uint pipeMode, uint maxInstances, uint outputBufferSize, uint inputBufferSize, uint timeoutInterval, IntPtr pipeSecurityDescriptor);

        bool ConnectNamedPipe(IntPtr namedPipeHandle, ref NativeOverlapped overlapped);

        bool DisconnectNamedPipe(IntPtr hHandle);

        bool ReadFile(IntPtr fileHandle, ref byte[] dataBuffer, uint nNumberOfBytesToRead, out uint lpNumberOfBytesRead, ref NativeOverlapped overlapped);

        bool WriteFile(IntPtr fileHandle, ref byte[] dataBuffer, uint nNumberOfBytesToWrite, out uint lpNumberOfBytesWritten, ref NativeOverlapped overlapped);

        bool PeekNamedPipe(IntPtr handle, ref byte[] buffer, uint nBufferSize, out uint bytesRead, out uint bytesAvail, out uint BytesLeftThisMessage);

        bool CallNamedPipe(string lpNamedPipeName, byte[] lpInBuffer, uint nInBufferSize, ref byte[] lpOutBuffer, uint nOutBufferSize, out uint lpBytesRead, uint nTimeOut);

        bool CloseHandle(IntPtr hObject);

        void SetLastError(uint dwErrCode);

        IntPtr GetModuleHandle(string lpModuleName);

        bool FlushFileBuffers(SafeFileHandle hFile);

		ToolHelpHandle CreateToolhelp32Snapshot(SnapshotFlags dwFlags, uint th32ProcessID);

		bool BitBlt(IntPtr hObject, int nXDest, int nYDest, int nWidth, int nHeight, IntPtr hObjSource, int nXSrc, int nYSrc, int dwRop);

        IntPtr CreateCompatibleDC(IntPtr hdc);

        bool DeleteDC(IntPtr hdc);

        IntPtr SelectObject(IntPtr hdc, IntPtr hgdiobj);

        bool DeleteObject(IntPtr hObject);

        IntPtr CreateRoundRectRgn(int nLeftRect, int nTopRect, int nRightRect, int nBottomRect, int nWidthEllipse, int nHeightEllipse);

        int SelectClipRgn(IntPtr hdc, IntPtr hrgn);

        int IsAppThemed();

        long DeleteUrlCacheEntry(string url);

        bool OpenProcessToken(IntPtr ProcessHandle, TOKEN_ACCESS DesiredAccess, out IntPtr TokenHandle);

		uint GetModuleFileNameEx(IntPtr hProcess, IntPtr hModule, out StringBuilder lpBaseName, ref int nSize);

        bool GetSaveFileName(ref OPENFILENAME lpofn);

        int CommDlgExtendedError();

    }

    public class NativeCall
    {
		private static INativeCall _instance;

		public static INativeCall Instance
		{
			get
			{
				if (_instance == null)
				{
					switch (Environment.OSVersion.Platform)
                    {
                        case PlatformID.Unix:
							_instance = new NativeCallUnix();
							break;

                        case PlatformID.MacOSX:
							_instance = new NativeCallOSX();
                            break;

                        case PlatformID.Win32NT:
							_instance = new NativeCallWin();
                            break;

                        default:
							_instance = new NativeCallDummy();
							break;
                    }
				}
				return _instance;
			}
		}
    }

}
